@@ -16,7 +16,7 @@ except Exception:
 # BCP check
 try:
     run(["bcp", "-v"], stdout=DEVNULL, stderr=DEVNULL, stdin=DEVNULL)
-except FileNotFoundError:
+except (FileNotFoundError, PermissionError):
     warnings.warn("BCP utility not installed or not found in PATH, bcpandas will not work!")
 
 del run, DEVNULL, warnings
