@@ -11,6 +11,7 @@ from pathlib import Path
 from textwrap import dedent
 from typing import Dict, List, Optional, Union
 from urllib.parse import quote_plus
+from re import sub
 
 import pandas as pd
 from pandas.io.sql import SQLDatabase, SQLTable
@@ -25,6 +26,7 @@ from bcpandas.constants import (
     BCPandasValueError,
     get_delimiter,
     get_quotechar,
+    sql_collation,
 )
 from bcpandas.utils import bcp, build_format_file, get_temp_file
 
@@ -71,6 +73,7 @@ class SqlCreds:
         self.server = server
         self.database = database
         self.port = port
+        self.odbc_kwargs = odbc_kwargs
 
         if driver_version is None:
             all_drivers: List[str] = pyodbc.drivers()
@@ -79,8 +82,10 @@ class SqlCreds:
             ]
             new_driver_version: int = max(int(v) for v in driver_candidates if v.isnumeric())
             self.driver = f"{{ODBC Driver {new_driver_version} for SQL Server}}"
+            self.driver_version = new_driver_version  # simplifies copy construction
         else:
             self.driver = f"{{ODBC Driver {driver_version} for SQL Server}}"
+            self.driver_version = driver_version  # simplifies copy construction
 
         # Append a comma for use in connection strings (optionally blank)
         if port:
@@ -95,22 +100,24 @@ class SqlCreds:
             self.username = username
             self.password = password
             self.with_krb_auth = False
-            db_url += f"UID={username};PWD={password}"
+            db_url += f"UID={username};PWD={password};"
         else:
             self.username = ""
             self.password = ""
             self.with_krb_auth = True
             db_url += "Trusted_Connection=yes;"
 
-        logger.info(f"Created creds:\t{self}")
+        self_msg = sub(r"password=\'.*\'", "password=[REDACTED]", str(self))
+        logger.info(f"Created creds:\t{self_msg}")
 
         # construct the engine for sqlalchemy
         if odbc_kwargs:
             db_url += ";".join(f"{k}={v}" for k, v in odbc_kwargs.items())
         conn_string = f"mssql+pyodbc:///?odbc_connect={quote_plus(db_url)}"
         self.engine = sa.engine.create_engine(conn_string)
+        engine_msg = sub("PWD%3D.*%3B", "PWD%3D[REDACTED]%3B", str(self.engine))
 
-        logger.info(f"Created engine for sqlalchemy:\t{self.engine}")
+        logger.info(f"Created engine for sqlalchemy:\t{engine_msg}")
 
     @classmethod
     def from_engine(cls, engine: sa.engine.base.Engine) -> "SqlCreds":
@@ -130,20 +137,30 @@ class SqlCreds:
         """
         try:
             # get the odbc url part from the engine, split by ';' delimiter
-            conn_url = engine.url.query["odbc_connect"].split(";")
-            # convert into dict
-            conn_dict = {x.split("=")[0].lower(): x.split("=")[1] for x in conn_url if "=" in x}
+            if "odbc_connect" in engine.url.query:
+                conn_url = engine.url.query["odbc_connect"].split(";")
+                # convert into dict
+                conn_dict = {
+                    x.split("=")[0].lower(): x.split("=")[1] for x in conn_url if "=" in x
+                }
 
-            if "," in conn_dict["server"]:
-                conn_dict["port"] = int(conn_dict["server"].split(",")[1])
-
-            sql_creds = cls(
-                server=conn_dict["server"].replace("tcp:", "").split(",")[0],
-                database=conn_dict["database"],
-                username=conn_dict.get("uid", None),
-                password=conn_dict.get("pwd", None),
-                port=conn_dict.get("port", None),
-            )
+                if "," in conn_dict["server"]:
+                    conn_dict["port"] = int(conn_dict["server"].split(",")[1])
+                sql_creds = cls(
+                    server=conn_dict["server"].replace("tcp:", "").split(",")[0],
+                    database=conn_dict["database"],
+                    username=conn_dict.get("uid", None),
+                    password=conn_dict.get("pwd", None),
+                    port=conn_dict.get("port", None),
+                )
+            elif "driver" in engine.url.query:
+                sql_creds = cls(
+                    server=engine.url.host,
+                    database=engine.url.database,
+                    username=engine.url.username,
+                    password=engine.url.password,
+                    port=engine.url.port,
+                )
 
             # add Engine object as attribute
             sql_creds.engine = engine
@@ -190,18 +207,19 @@ def _create_table(
 ):
     """use pandas' own code to create the table and schema"""
 
-    sql_db = SQLDatabase(engine=creds.engine, schema=schema)
-    table = SQLTable(
-        table_name,
-        sql_db,
-        frame=df,
-        index=False,  # already set as new col earlier if index=True
-        if_exists=if_exists,
-        index_label=None,
-        schema=schema,
-        dtype=dtype,
-    )
-    table.create()
+    with creds.engine.begin() as conn:
+        sql_db = SQLDatabase(conn, schema=schema)
+        table = SQLTable(
+            table_name,
+            sql_db,
+            frame=df,
+            index=False,  # already set as new col earlier if index=True
+            if_exists=if_exists,
+            index_label=None,
+            schema=schema,
+            dtype=dtype,
+        )
+        table.create()
 
 
 def _handle_cols_for_append(
@@ -336,6 +354,8 @@ def to_sql(
     quotechar: Optional[str] = None,
     encoding: Optional[str] = None,
     work_directory: Optional[Path] = None,
+    collation: str = sql_collation,
+    identity_insert: bool = False,
 ):
     """
     Writes the pandas DataFrame to a SQL table or view.
@@ -396,6 +416,8 @@ def to_sql(
     work_directory: pathlib.Path, default None
         Optional directory where temporary files are written to. If not provided, defaults to the
         system-default for temporary files.
+    identity_insert: bool, default False
+        Specifies that identity value or values in the imported data file are to be used for the identity column.
 
     Notes
     -----
@@ -408,9 +430,8 @@ def to_sql(
 
     _validate_args(df=df, sql_type=sql_type, if_exists=if_exists, batch_size=batch_size)
 
-    # TODO diff way to implement? could be big performance hit with big dataframe
     if index:
-        df = df.copy(deep=True).reset_index()
+        df = df.reset_index()
 
     delim = get_delimiter(df) if delimiter is None else delimiter
     _quotechar = get_quotechar(df) if quotechar is None else quotechar
@@ -425,7 +446,7 @@ def to_sql(
         index=False,  # already set as new col earlier if index=True
         quoting=csv.QUOTE_MINIMAL,  # pandas default
         quotechar=_quotechar,
-        line_terminator=NEWLINE,
+        lineterminator=NEWLINE,
         doublequote=True,
         escapechar=None,  # not needed, as using doublequote
         encoding=encoding,
@@ -448,7 +469,9 @@ def to_sql(
         if_exists=if_exists,
     )
 
-    fmt_file_txt = build_format_file(df=df, delimiter=delim, db_cols_order=cols_dict)
+    fmt_file_txt = build_format_file(
+        df=df, delimiter=delim, db_cols_order=cols_dict, collation=collation
+    )
     with open(fmt_file_path, "w") as ff:
         ff.write(fmt_file_txt)
     logger.debug(f"Created BCP format file at {fmt_file_path}")
@@ -479,6 +502,7 @@ def to_sql(
             batch_size=batch_size,
             use_tablock=use_tablock,
             bcp_path=bcp_path,
+            identity_insert=identity_insert,
         )
     finally:
         if not debug:
