@@ -5,13 +5,15 @@ Created on Sat Aug  3 23:07:15 2019
 """
 
 import logging
+import sys
 from pathlib import Path
 import random
 import shlex
 import string
-from subprocess import PIPE, Popen
+from subprocess import PIPE, STDOUT, Popen
 import tempfile
-from typing import Dict, List, Optional, Union
+from typing import Dict, List, Optional, Tuple, Union
+from re import sub
 
 import pandas as pd
 
@@ -49,6 +51,7 @@ def bcp(
     col_delimiter: Optional[str] = None,
     row_terminator: Optional[str] = None,
     bcp_path: Optional[Union[str, Path]] = None,
+    identity_insert: bool = False,
 ):
     """
     See https://docs.microsoft.com/en-us/sql/tools/bcp-utility
@@ -70,6 +73,12 @@ def bcp(
         auth = ["-T"]
     else:
         auth = ["-U", quote_this(creds.username), "-P", quote_this(creds.password)]
+    if creds.odbc_kwargs:
+        kwargs = {k.lower(): v for k, v in creds.odbc_kwargs.items()}
+        false_values = ("n", "no", "f", "false", "off", "0")
+
+        if sys.platform != "win32" and "encrypt" in kwargs:
+            auth += [f"-Y{'m' if kwargs['encrypt'] not in false_values else 'o'}"]
 
     # prepare SQL item string
     if sql_type == QUERY:
@@ -102,6 +111,9 @@ def bcp(
     if use_tablock:
         bcp_command += ["-h", "TABLOCK"]
 
+    if identity_insert:
+        bcp_command += ["-E"]
+
     # formats
     if direc == IN and format_file_path is not None:
         bcp_command += ["-f", str(format_file_path)]
@@ -117,11 +129,15 @@ def bcp(
         ]
 
     # execute
-    bcp_command_log = [c if c != creds.password else "[REDACTED]" for c in bcp_command]
-    logger.info(f"Executing BCP command now... \nBCP command is: {bcp_command_log}")
-    ret_code = run_cmd(bcp_command, print_output=print_output)
-    if ret_code:
-        raise BCPandasException(f"Bcp command failed with exit code {ret_code}")
+    bcp_command_log = ", ".join(bcp_command)
+    bcp_command_log_msg = sub(r"-P,\s.*,", "-P, [REDACTED],", bcp_command_log)
+    logger.info(f"Executing BCP command now... \nBCP command is: {bcp_command_log_msg}")
+    ret_code, output = run_cmd(bcp_command, print_output=print_output)
+    if ret_code != 0:
+        raise BCPandasException(
+            f"Bcp command failed with exit code {ret_code}",
+            details=[line for line in output if line.startswith("Error =")],
+        )
 
 
 def get_temp_file(directory: Optional[Path] = None) -> Path:
@@ -148,7 +164,10 @@ def _escape(input_string: str) -> str:
 
 
 def build_format_file(
-    df: pd.DataFrame, delimiter: str, db_cols_order: Optional[Dict[str, int]] = None
+    df: pd.DataFrame,
+    delimiter: str,
+    db_cols_order: Optional[Dict[str, int]] = None,
+    collation: str = sql_collation,
 ) -> str:
     """
     Creates the non-xml SQL format file. Puts 4 spaces between each section.
@@ -167,6 +186,8 @@ def build_format_file(
         Maps existing columns in the database to their ordinal position, i.e. the order of the columns in the db table.
         1-indexed, so the first columns is 1, second is 2, etc.
         Only needed if the order of the columns in the dataframe doesn't match the database.
+    collation: str, optional
+        Collation to be used in the format file. The default value is 'SQL_Latin1_General_CP1_CI_AS'
 
     Returns
     -------
@@ -187,8 +208,10 @@ def build_format_file(
                 str(
                     col_num if not db_cols_order else db_cols_order[str(col_name)]
                 ),  # Server column order
-                str(col_name),  # Server column name, optional as long as not blank
-                sql_collation,  # Column collation
+                str(col_name).replace(
+                    " ", r"\s"
+                ),  # Server column name, optional as long as not blank
+                collation,  # Column collation
                 "\n",
             ]
         )
@@ -215,7 +238,7 @@ def quote_this(this: str, skip: bool = False) -> str:
         return this
 
 
-def run_cmd(cmd: List[str], *, print_output: bool) -> int:
+def run_cmd(cmd: List[str], *, print_output: bool) -> Tuple[int, List[str]]:
     """
     Runs the given command.
 
@@ -232,7 +255,7 @@ def run_cmd(cmd: List[str], *, print_output: bool) -> int:
 
     Returns
     -------
-    The exit code of the command
+    The exit code of the command and all of its output.
     """
     if IS_WIN32:
         with_shell = False
@@ -242,23 +265,20 @@ def run_cmd(cmd: List[str], *, print_output: bool) -> int:
     proc = Popen(
         cmd,
         stdout=PIPE,
-        stderr=PIPE,
+        stderr=STDOUT,
         encoding="utf-8",
         errors="utf-8",
         shell=with_shell,
     )
-    # live stream STDOUT
+    stdout = []
+    # live stream STDOUT and STDERR
     while True:
         outs = proc.stdout.readline()  # type: ignore[union-attr]
         if outs:
             if print_output:
                 print(outs, end="")
             logger.info(outs)
+            stdout.append(outs)
         if proc.poll() is not None and outs == "":
             break
-    errs = proc.stderr.readlines()  # type: ignore[union-attr]
-    if errs:
-        if print_output:
-            print(errs, end="")
-        logger.error(errs)
-    return proc.returncode
+    return proc.returncode, stdout
